{%- extends "epub/epub-cover.html" %}

{% block content %}
  <div class="epub-cover">
    <h1>Django Documentation</h1>
    <h2><em>Release {{ release }}</em></h2>
    <h3>{{ copyright }}</h3>
    <p>{{ last_updated }}</p>
  </div>
{% endblock %}
