{% extends "admin/change_list_object_tools.html" %}
{% load i18n admin_urls %}

{% block object-tools-items %}
<li><a href="#" id="change-list-export" class="override-change_list_object_tools change-list-object-tools-item">{% translate "Export" %}</a></li>
{{ block.super }}
{% endblock %}
